@@ -23,6 +23,14 @@
 # include "compat/stdbool.h"
 #endif /* HAVE_STDBOOL_H */
 
+#ifndef TIME_T_MAX
+# if SIZEOF_TIME_T == 8
+#  define TIME_T_MAX	LLONG_MAX
+# else
+#  define TIME_T_MAX	INT_MAX
+# endif
+#endif
+
 /*
  * Macros for operating on struct timeval.
  */
@@ -170,6 +178,8 @@ __dso_public char *sudo_gethostname_v1(void);
 #define sudo_gethostname() sudo_gethostname_v1()
 
 /* gettime.c */
+__dso_public int sudo_gettime_awake_v1(struct timespec *ts);
+#define sudo_gettime_awake(_a) sudo_gettime_awake_v1((_a))
 __dso_public int sudo_gettime_mono_v1(struct timespec *ts);
 #define sudo_gettime_mono(_a) sudo_gettime_mono_v1((_a))
 __dso_public int sudo_gettime_real_v1(struct timespec *ts);
