@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2004-2005, 2007, 2010, 2012-2015
- *	Todd C. Miller <Todd.Miller@courtesan.com>
+ *	Todd C. Miller <Todd.Miller@sudo.ws>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -15,6 +15,11 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+/*
+ * This is an open source non-commercial project. Dear PVS-Studio, please check it.
+ * PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
+ */
+
 #include <config.h>
 
 #ifndef HAVE_CLOSEFROM
@@ -38,15 +43,11 @@
 # define _POSIX_OPEN_MAX	20
 #endif
 
-#if defined(HAVE_FCNTL_CLOSEM) && !defined(HAVE_DIRFD)
-# define sudo_closefrom	closefrom_fallback
-#endif
-
 /*
  * Close all file descriptors greater than or equal to lowfd.
  * This is the expensive (fallback) method.
  */
-void
+static void
 closefrom_fallback(int lowfd)
 {
     long fd, maxfd;
@@ -126,5 +127,12 @@ sudo_closefrom(int lowfd)
     } else
 	closefrom_fallback(lowfd);
 }
+#else
+void
+sudo_closefrom(int lowfd)
+{
+    closefrom_fallback(lowfd);
+}
 #endif /* HAVE_FCNTL_CLOSEM */
+
 #endif /* HAVE_CLOSEFROM */
