@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-1996, 1998-2017 Todd C. Miller <Todd.Miller@sudo.ws>
+ * Copyright (c) 1993-1996, 1998-2018 Todd C. Miller <Todd.Miller@sudo.ws>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -18,6 +18,11 @@
  * Materiel Command, USAF, under agreement number F39502-99-1-0512.
  */
 
+/*
+ * This is an open source non-commercial project. Dear PVS-Studio, please check it.
+ * PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
+ */
+
 #ifdef __TANDEM
 # include <floss.h>
 #endif
@@ -59,6 +64,7 @@
 #include <ctype.h>
 
 #include "sudoers.h"
+#include "parse.h"
 #include "auth/sudo_auth.h"
 
 #ifndef HAVE_GETADDRINFO
@@ -84,7 +90,6 @@ static bool tty_present(void);
  */
 struct sudo_user sudo_user;
 struct passwd *list_pw;
-int long_list;
 uid_t timestamp_uid;
 gid_t timestamp_gid;
 #ifdef HAVE_BSD_AUTH_H
@@ -189,16 +194,19 @@ sudoers_policy_init(void *info, char * const envp[])
      */
     sudoers_setlocale(SUDOERS_LOCALE_SUDOERS, &oldlocale);
     sudo_warn_set_locale_func(sudoers_warn_setlocale);
+    init_parser(sudoers_file, false);
     TAILQ_FOREACH_SAFE(nss, snl, entries, nss_next) {
-        if (nss->open(nss) == 0 && nss->parse(nss) == 0) {
-            sources++;
-            if (nss->setdefs(nss) != 0) {
-                log_warningx(SLOG_SEND_MAIL|SLOG_NO_STDERR,
-		    N_("problem with defaults entries"));
-	    }
-        } else {
+	if (nss->open(nss) == -1 || (nss->parse_tree = nss->parse(nss)) == NULL) {
 	    TAILQ_REMOVE(snl, nss, entries);
-        }
+	    continue;
+	}
+
+	sources++;
+	if (nss->getdefs(nss) == -1 || !update_defaults(nss->parse_tree, NULL,
+	    SETDEF_GENERIC|SETDEF_HOST|SETDEF_USER|SETDEF_RUNAS, false)) {
+	    log_warningx(SLOG_SEND_MAIL|SLOG_NO_STDERR,
+		N_("problem with defaults entries"));
+	}
     }
     if (sources == 0) {
 	sudo_warnx(U_("no valid sudoers sources found, quitting"));
@@ -222,13 +230,12 @@ cleanup:
 
 int
 sudoers_policy_main(int argc, char * const argv[], int pwflag, char *env_add[],
-    void *closure)
+    bool verbose, void *closure)
 {
     char **edit_argv = NULL;
     char *iolog_path = NULL;
     mode_t cmnd_umask = ACCESSPERMS;
     struct sudo_nss *nss;
-    int nopass = -1;
     int cmnd_status = -1, oldlocale, validated;
     int ret = -1;
     debug_decl(sudoers_policy_main, SUDOERS_DEBUG_PLUGIN)
@@ -240,12 +247,12 @@ sudoers_policy_main(int argc, char * const argv[], int pwflag, char *env_add[],
     /* Is root even allowed to run sudo? */
     if (user_uid == 0 && !def_root_sudo) {
 	/* Not an audit event. */
-        sudo_warnx(U_("sudoers specifies that root is not allowed to sudo"));
-        goto bad;
+	sudo_warnx(U_("sudoers specifies that root is not allowed to sudo"));
+	goto bad;
     }    
 
     if (!set_perms(PERM_INITIAL))
-        goto bad;
+	goto bad;
 
     /* Environment variables specified on the command line. */
     if (env_add != NULL && env_add[0] != NULL)
@@ -272,7 +279,8 @@ sudoers_policy_main(int argc, char * const argv[], int pwflag, char *env_add[],
 	    sudo_warnx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
 	    goto done;
 	}
-	memcpy(++NewArgv, argv, argc * sizeof(char *));
+	NewArgv++;	/* reserve an extra slot for --login */
+	memcpy(NewArgv, argv, argc * sizeof(char *));
 	NewArgv[NewArgc] = NULL;
 	if (ISSET(sudo_mode, MODE_LOGIN_SHELL) && runas_pw != NULL) {
 	    NewArgv[0] = strdup(runas_pw->pw_shell);
@@ -307,54 +315,12 @@ sudoers_policy_main(int argc, char * const argv[], int pwflag, char *env_add[],
      * Check sudoers sources, using the locale specified in sudoers.
      */
     sudoers_setlocale(SUDOERS_LOCALE_SUDOERS, &oldlocale);
-    validated = FLAG_NO_USER | FLAG_NO_HOST;
-    TAILQ_FOREACH(nss, snl, entries) {
-	validated = nss->lookup(nss, validated, pwflag);
-
-	/*
-	 * The NOPASSWD tag needs special handling among all sources
-	 * in -l or -v mode.
-	 */
-	if (pwflag) {
-	    enum def_tuple pwcheck =
-		(pwflag == -1) ? never : sudo_defs_table[pwflag].sd_un.tuple;
-	    switch (pwcheck) {
-	    case all:
-		if (!ISSET(validated, FLAG_NOPASSWD))
-		    nopass = false;
-		else if (nopass == -1)
-		    nopass = true;
-		break;
-	    case any:
-		if (ISSET(validated, FLAG_NOPASSWD))
-		    nopass = true;
-		break;
-	    case never:
-		nopass = true;
-		break;
-	    case always:
-		nopass = false;
-		break;
-	    default:
-		break;
-	    }
-	}
-
-	if (ISSET(validated, VALIDATE_ERROR)) {
-	    /* The lookup function should have printed an error. */
-	    goto done;
-	} else if (ISSET(validated, VALIDATE_SUCCESS)) {
-	    /* Handle [SUCCESS=return] */
-	    if (nss->ret_if_found)
-		break;
-	} else {
-	    /* Handle [NOTFOUND=return] */
-	    if (nss->ret_if_notfound)
-		break;
-	}
+    validated = sudoers_lookup(snl, sudo_user.pw, FLAG_NO_USER | FLAG_NO_HOST,
+	pwflag);
+    if (ISSET(validated, VALIDATE_ERROR)) {
+	/* The lookup function should have printed an error. */
+	goto done;
     }
-    if (pwflag && nopass == true)
-	def_authenticate = false;
 
     /* Restore user's locale. */
     sudoers_setlocale(oldlocale, NULL);
@@ -524,7 +490,7 @@ sudoers_policy_main(int argc, char * const argv[], int pwflag, char *env_add[],
 	    ret = display_cmnd(snl, list_pw ? list_pw : sudo_user.pw);
 	    break;
 	case MODE_LIST:
-	    ret = display_privs(snl, list_pw ? list_pw : sudo_user.pw);
+	    ret = display_privs(snl, list_pw ? list_pw : sudo_user.pw, verbose);
 	    break;
 	case MODE_VALIDATE:
 	    /* Nothing to do. */
@@ -546,6 +512,7 @@ sudoers_policy_main(int argc, char * const argv[], int pwflag, char *env_add[],
     }
     if (def_group_plugin)
 	group_plugin_unload();
+    init_parser(NULL, false);
 
     if (ISSET(sudo_mode, (MODE_VALIDATE|MODE_CHECK|MODE_LIST))) {
 	/* ret already set appropriately */
@@ -801,8 +768,9 @@ init_vars(char * const envp[])
 static int
 set_cmnd(void)
 {
-    int ret = FOUND;
+    struct sudo_nss *nss;
     char *path = user_path;
+    int ret = FOUND;
     debug_decl(set_cmnd, SUDOERS_DEBUG_PLUGIN)
 
     /* Allocate user_stat for find_path() and match functions. */
@@ -890,9 +858,11 @@ set_cmnd(void)
     else
 	user_base = user_cmnd;
 
-    if (!update_defaults(SETDEF_CMND, false)) {
-	log_warningx(SLOG_SEND_MAIL|SLOG_NO_STDERR,
-	    N_("problem with defaults entries"));
+    TAILQ_FOREACH(nss, snl, entries) {
+	if (!update_defaults(nss->parse_tree, NULL, SETDEF_CMND, false)) {
+	    log_warningx(SLOG_SEND_MAIL|SLOG_NO_STDERR,
+		N_("problem with defaults entries"));
+	}
     }
 
     debug_return_int(ret);
@@ -1040,29 +1010,30 @@ set_loginclass(struct passwd *pw)
  * Returns true on success, setting longp and shortp.
  * Returns false on failure, longp and shortp are unchanged.
  */
-static bool
+static int
 resolve_host(const char *host, char **longp, char **shortp)
 {
     struct addrinfo *res0, hint;
     char *cp, *lname, *sname;
+    int ret;
     debug_decl(resolve_host, SUDOERS_DEBUG_PLUGIN)
 
     memset(&hint, 0, sizeof(hint));
     hint.ai_family = PF_UNSPEC;
     hint.ai_flags = AI_FQDN;
 
-    if (getaddrinfo(host, NULL, &hint, &res0) != 0)
-	debug_return_bool(false);
+    if ((ret = getaddrinfo(host, NULL, &hint, &res0)) != 0)
+	debug_return_int(ret);
     if ((lname = strdup(res0->ai_canonname)) == NULL) {
 	freeaddrinfo(res0);
-	debug_return_bool(false);
+	debug_return_int(EAI_MEMORY);
     }
     if ((cp = strchr(lname, '.')) != NULL) {
 	sname = strndup(lname, (size_t)(cp - lname));
 	if (sname == NULL) {
 	    free(lname);
 	    freeaddrinfo(res0);
-	    debug_return_bool(false);
+	    debug_return_int(EAI_MEMORY);
 	}
     } else {
 	sname = lname;
@@ -1071,7 +1042,7 @@ resolve_host(const char *host, char **longp, char **shortp)
     *longp = lname;
     *shortp = sname;
 
-    debug_return_bool(true);
+    debug_return_bool(0);
 }
 
 /*
@@ -1093,9 +1064,10 @@ cb_fqdn(const union sudo_defs_val *sd_un)
     remote = strcmp(user_runhost, user_host) != 0;
 
     /* First resolve user_host, setting user_host and user_shost. */
-    if (!resolve_host(user_host, &lhost, &shost)) {
-	if (!resolve_host(user_runhost, &lhost, &shost)) {
-	    log_warning(SLOG_SEND_MAIL|SLOG_RAW_MSG,
+    if (resolve_host(user_host, &lhost, &shost) != 0) {
+	int rc = resolve_host(user_runhost, &lhost, &shost);
+	if (rc != 0) {
+	    gai_log_warning(SLOG_SEND_MAIL|SLOG_RAW_MSG, rc,
 		N_("unable to resolve host %s"), user_host);
 	    debug_return_bool(false);
 	}
@@ -1298,7 +1270,7 @@ create_admin_success_flag(void)
 static bool
 tty_present(void)
 {
-#if defined(HAVE_STRUCT_KINFO_PROC2_P_TDEV) || defined(HAVE_STRUCT_KINFO_PROC_P_TDEV) || defined(HAVE_STRUCT_KINFO_PROC_KI_TDEV) || defined(HAVE_STRUCT_KINFO_PROC_KP_EPROC_E_TDEV) || defined(HAVE_STRUCT_PSINFO_PR_TTYDEV) || defined(HAVE_PSTAT_GETPROC) || defined(__linux__)
+#if defined(HAVE_KINFO_PROC2_NETBSD) || defined(HAVE_KINFO_PROC_OPENBSD) || defined(HAVE_KINFO_PROC_FREEBSD) || defined(HAVE_KINFO_PROC_44BSD) || defined(HAVE_STRUCT_PSINFO_PR_TTYDEV) || defined(HAVE_PSTAT_GETPROC) || defined(__linux__)
     return user_ttypath != NULL;
 #else
     int fd = open(_PATH_TTY, O_RDWR);
