@@ -14,12 +14,16 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+/*
+ * This is an open source non-commercial project. Dear PVS-Studio, please check it.
+ * PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
+ */
+
 #include <config.h>
 
 #include <sys/types.h>
 #include <sys/uio.h>
 #include <sys/stat.h>
-#include <sys/time.h>
 #include <sys/wait.h>
 #include <sys/ioctl.h>
 #include <stdio.h>
@@ -49,9 +53,7 @@
 #endif /* HAVE_STDBOOL_H */
 #include <regex.h>
 #include <signal.h>
-#ifdef HAVE_ZLIB_H
-# include <zlib.h>
-#endif
+#include <time.h>
 
 #include <pathnames.h>
 
@@ -61,6 +63,7 @@
 #include "sudo_fatal.h"
 #include "logging.h"
 #include "iolog.h"
+#include "iolog_files.h"
 #include "sudo_queue.h"
 #include "sudo_plugin.h"
 #include "sudo_conf.h"
@@ -74,21 +77,6 @@
 # include "compat/getopt.h"
 #endif /* HAVE_GETOPT_LONG */
 
-/*
- * Info present in the I/O log file
- */
-struct log_info {
-    char *cwd;
-    char *user;
-    char *runas_user;
-    char *runas_group;
-    char *tty;
-    char *cmd;
-    time_t tstamp;
-    int rows;
-    int cols;
-};
-
 struct replay_closure {
     struct sudo_event_base *evbase;
     struct sudo_event *delay_ev;
@@ -99,19 +87,9 @@ struct replay_closure {
     struct sudo_event *sigquit_ev;
     struct sudo_event *sigterm_ev;
     struct sudo_event *sigtstp_ev;
-    struct timing_closure {
-	const char *decimal;
-	struct timeval *max_delay;
-	int idx;
-	union {
-	    struct {
-		int rows;
-		int cols;
-	    } winsize;
-	    size_t nbytes; // XXX
-	} u;
-    } timing;
+    struct timing_closure timing;
     bool interactive;
+    bool suspend_wait;
     struct io_buffer {
 	unsigned int len; /* buffer length (how much produced) */
 	unsigned int off; /* write position (how much already consumed) */
@@ -155,8 +133,6 @@ struct search_node {
 
 static struct search_node_list search_expr = STAILQ_HEAD_INITIALIZER(search_expr);
 
-static int timing_idx_adj;
-
 static double speed_factor = 1.0;
 
 static const char *session_dir = _PATH_SUDO_IO_LOGDIR;
@@ -167,7 +143,7 @@ static int terminal_rows, terminal_cols;
 
 static int ttyfd = -1;
 
-static const char short_opts[] =  "d:f:hlm:nRs:V";
+static const char short_opts[] =  "d:f:hlm:nRSs:V";
 static struct option long_opts[] = {
     { "directory",	required_argument,	NULL,	'd' },
     { "filter",		required_argument,	NULL,	'f' },
@@ -176,24 +152,22 @@ static struct option long_opts[] = {
     { "max-wait",	required_argument,	NULL,	'm' },
     { "non-interactive", no_argument,		NULL,	'n' },
     { "no-resize",	no_argument,		NULL,	'R' },
+    { "suspend-wait",	no_argument,		NULL,	'S' },
     { "speed",		required_argument,	NULL,	's' },
     { "version",	no_argument,		NULL,	'V' },
     { NULL,		no_argument,		NULL,	'\0' },
 };
 
-/* XXX move to separate header? */
+/* XXX move to separate header? (currently in sudoers.h) */
 extern char *get_timestr(time_t, int);
 extern time_t get_date(char *);
 
 static int list_sessions(int, char **, const char *, const char *, const char *);
 static int open_io_fd(char *path, int len, struct io_log_file *iol);
 static int parse_expr(struct search_node_list *, char **, bool);
-static bool parse_timing(const char *buf, double *seconds, struct timing_closure *timing);
-static struct log_info *parse_logfile(char *logfile);
 static void read_keyboard(int fd, int what, void *v);
-static void free_log_info(struct log_info *li);
 static void help(void) __attribute__((__noreturn__));
-static int replay_session(struct timeval *max_wait, const char *decimal, bool interactive);
+static int replay_session(struct timespec *max_wait, const char *decimal, bool interactive, bool suspend_wait);
 static void sudoreplay_cleanup(void);
 static void usage(int);
 static void write_output(int fd, int what, void *v);
@@ -219,13 +193,13 @@ __dso_public int main(int argc, char *argv[]);
 int
 main(int argc, char *argv[])
 {
-    int ch, idx, plen, exitcode = 0;
+    int ch, i, plen, exitcode = 0;
     bool def_filter = true, listonly = false;
-    bool interactive = true, resize = true;
+    bool interactive = true, suspend_wait = false, resize = true;
     const char *decimal, *id, *user = NULL, *pattern = NULL, *tty = NULL;
     char *cp, *ep, path[PATH_MAX];
     struct log_info *li;
-    struct timeval max_delay_storage, *max_delay = NULL;
+    struct timespec max_delay_storage, *max_delay = NULL;
     double dval;
     debug_decl(main, SUDO_DEBUG_MAIN)
 
@@ -286,11 +260,11 @@ main(int argc, char *argv[])
 	    if (*ep != '\0' || errno != 0)
 		sudo_fatalx(U_("invalid max wait: %s"), optarg);
 	    if (dval <= 0.0) {
-		sudo_timevalclear(&max_delay_storage);
+		sudo_timespecclear(&max_delay_storage);
 	    } else {
 		max_delay_storage.tv_sec = dval;
-		max_delay_storage.tv_usec =
-		    (dval - max_delay_storage.tv_sec) * 1000000.0;
+		max_delay_storage.tv_nsec =
+		    (dval - max_delay_storage.tv_sec) * 1000000000.0;
 	    }
 	    max_delay = &max_delay_storage;
 	    break;
@@ -300,6 +274,9 @@ main(int argc, char *argv[])
 	case 'R':
 	    resize = false;
 	    break;
+	case 'S':
+	    suspend_wait = true;
+	    break;
 	case 's':
 	    errno = 0;
 	    speed_factor = strtod(optarg, &ep);
@@ -351,8 +328,8 @@ main(int argc, char *argv[])
     plen -= 7;
 
     /* Open files for replay, applying replay filter for the -f flag. */
-    for (idx = 0; idx < IOFD_MAX; idx++) {
-	if (open_io_fd(path, plen, &io_log_files[idx]) == -1)
+    for (i = 0; i < IOFD_MAX; i++) {
+	if (open_io_fd(path, plen, &io_log_files[i]) == -1)
 	    sudo_fatal(U_("unable to open %s"), path);
     }
 
@@ -375,7 +352,7 @@ main(int argc, char *argv[])
     li = NULL;
 
     /* Replay session corresponding to io_log_files[]. */
-    exitcode = replay_session(max_delay, decimal, interactive);
+    exitcode = replay_session(max_delay, decimal, interactive, suspend_wait);
 
     restore_terminal_size();
     sudo_term_restore(ttyfd, true);
@@ -389,7 +366,7 @@ done:
  * Return 0 for EOF or -1 on error.
  */
 static ssize_t
-io_log_read(int idx, char *buf, size_t nbytes)
+io_log_read(union io_fd ifd, char *buf, size_t nbytes)
 {
     ssize_t nread;
     debug_decl(io_log_read, SUDO_DEBUG_UTIL)
@@ -399,39 +376,39 @@ io_log_read(int idx, char *buf, size_t nbytes)
 	debug_return_ssize_t(-1);
     }
 #ifdef HAVE_ZLIB_H
-    nread = gzread(io_log_files[idx].fd.g, buf, nbytes);
+    nread = gzread(ifd.g, buf, nbytes);
 #else
-    nread = (ssize_t)fread(buf, 1, nbytes, io_log_files[idx].fd.f);
-    if (nread == 0 && ferror(io_log_files[idx].fd.f))
+    nread = (ssize_t)fread(buf, 1, nbytes, ifd.f);
+    if (nread == 0 && ferror(ifd.f))
 	nread = -1;
 #endif
     debug_return_ssize_t(nread);
 }
 
 static int
-io_log_eof(int idx)
+io_log_eof(union io_fd ifd)
 {
     int ret;
     debug_decl(io_log_eof, SUDO_DEBUG_UTIL)
 
 #ifdef HAVE_ZLIB_H
-    ret = gzeof(io_log_files[idx].fd.g);
+    ret = gzeof(ifd.g);
 #else
-    ret = feof(io_log_files[idx].fd.f);
+    ret = feof(ifd.f);
 #endif
     debug_return_int(ret);
 }
 
 static char *
-io_log_gets(int idx, char *buf, size_t nbytes)
+io_log_gets(union io_fd ifd, char *buf, size_t nbytes)
 {
     char *str;
     debug_decl(io_log_gets, SUDO_DEBUG_UTIL)
 
 #ifdef HAVE_ZLIB_H
-    str = gzgets(io_log_files[idx].fd.g, buf, nbytes);
+    str = gzgets(ifd.g, buf, nbytes);
 #else
-    str = fgets(buf, nbytes, io_log_files[idx].fd.f);
+    str = fgets(buf, nbytes, ifd.f);
 #endif
     debug_return_str(str);
 }
@@ -465,7 +442,7 @@ struct getsize_closure {
     int state;
     const char *cp;
     struct sudo_event *ev;
-    struct timeval timeout;
+    struct timespec timeout;
 };
 
 /* getsize states */
@@ -594,15 +571,15 @@ xterm_get_size(int *new_rows, int *new_cols)
     gc.nums_maxdepth = 1;
     gc.cp = getsize_response;
     gc.timeout.tv_sec = 10;
-    gc.timeout.tv_usec = 0;
+    gc.timeout.tv_nsec = 0;
 
     /* Setup an event for reading the terminal size */
     evbase = sudo_ev_base_alloc();
     if (evbase == NULL)
-	sudo_fatal(NULL);
+	sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
     gc.ev = sudo_ev_alloc(ttyfd, SUDO_EV_READ, getsize_cb, &gc);
     if (gc.ev == NULL)
-        sudo_fatal(NULL);
+	sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
 
     /* Read back terminal size response */
     if (sudo_ev_add(evbase, gc.ev, &gc.timeout, false) == -1)
@@ -773,43 +750,33 @@ restore_terminal_size(void)
 static int
 read_timing_record(struct replay_closure *closure)
 {
-    struct timeval timeout;
+    struct timespec timeout;
     char buf[LINE_MAX];
-    double delay;
     debug_decl(read_timing_record, SUDO_DEBUG_UTIL)
 
     /* Read next record from timing file. */
-    if (io_log_gets(IOFD_TIMING, buf, sizeof(buf)) == NULL) {
+    if (io_log_gets(io_log_files[IOFD_TIMING].fd, buf, sizeof(buf)) == NULL) {
 	/* EOF or error reading timing file, we are done. */
-	debug_return_int(io_log_eof(IOFD_TIMING) ? 1 : -1);
+	debug_return_int(io_log_eof(io_log_files[IOFD_TIMING].fd) ? 1 : -1);
     }
 
     /* Parse timing file record. */
     buf[strcspn(buf, "\n")] = '\0';
-    if (!parse_timing(buf, &delay, &closure->timing))
+    if (!parse_timing(buf, &timeout, &closure->timing))
 	sudo_fatalx(U_("invalid timing file line: %s"), buf);
 
     /* Record number bytes to read. */
     /* XXX - remove timing->nbytes? */
-    if (closure->timing.idx != IOFD_TIMING) {
+    if (closure->timing.event != IO_EVENT_WINSIZE &&
+	closure->timing.event != IO_EVENT_SUSPEND) {
     	closure->iobuf.len = 0;
     	closure->iobuf.off = 0;
     	closure->iobuf.lastc = '\0';
     	closure->iobuf.toread = closure->timing.u.nbytes;
     }
 
-    /* Adjust delay using speed factor. */
-    delay /= speed_factor;
-
-    /* Convert delay to a timeval. */
-    timeout.tv_sec = delay;
-    timeout.tv_usec = (delay - timeout.tv_sec) * 1000000.0;
-
-    /* Clamp timeout to max delay. */
-    if (closure->timing.max_delay != NULL) {
-	if (sudo_timevalcmp(&timeout, closure->timing.max_delay, >))
-	    timeout = *closure->timing.max_delay;
-    }
+    /* Adjust delay using speed factor and max_delay. */
+    adjust_delay(&timeout, closure->timing.max_delay, speed_factor);
 
     /* Schedule the delay event. */
     if (sudo_ev_add(closure->evbase, closure->delay_ev, &timeout, false) == -1)
@@ -827,9 +794,15 @@ next_timing_record(struct replay_closure *closure)
 {
     debug_decl(next_timing_record, SUDO_DEBUG_UTIL)
 
+again:
     switch (read_timing_record(closure)) {
     case 0:
 	/* success */
+	if (closure->timing.event == IO_EVENT_SUSPEND &&
+	    closure->timing.u.signo == SIGCONT && !closure->suspend_wait) {
+	    /* Ignore time spent suspended. */
+	    goto again;
+	}
 	break;
     case 1:
 	/* EOF */
@@ -848,25 +821,24 @@ fill_iobuf(struct replay_closure *closure)
 {
     const size_t space = sizeof(closure->iobuf.buf) - closure->iobuf.len;
     const struct timing_closure *timing = &closure->timing;
-    ssize_t nread;
-    size_t len;
     debug_decl(fill_iobuf, SUDO_DEBUG_UTIL)
 
     if (closure->iobuf.toread != 0 && space != 0) {
-	len = closure->iobuf.toread < space ? closure->iobuf.toread : space;
-	nread = io_log_read(timing->idx,
+	const size_t len =
+	    closure->iobuf.toread < space ? closure->iobuf.toread : space;
+	ssize_t nread = io_log_read(timing->fd,
 	    closure->iobuf.buf + closure->iobuf.off, len);
 	if (nread <= 0) {
 	    if (nread == 0) {
 		sudo_debug_printf(SUDO_DEBUG_ERROR|SUDO_DEBUG_LINENO,
 		    "%s: premature EOF, expected %u bytes",
-		    io_log_files[timing->idx].suffix, closure->iobuf.toread);
+		    io_log_files[timing->event].suffix, closure->iobuf.toread);
 	    } else {
 		sudo_debug_printf(SUDO_DEBUG_ERROR|SUDO_DEBUG_ERRNO|SUDO_DEBUG_LINENO,
-		    "%s: read error", io_log_files[timing->idx].suffix);
+		    "%s: read error", io_log_files[timing->event].suffix);
 	    }
 	    sudo_warnx(U_("unable to read %s"),
-		io_log_files[timing->idx].suffix);
+		io_log_files[timing->event].suffix);
 	    debug_return_bool(false);
 	}
 	closure->iobuf.toread -= nread;
@@ -885,26 +857,44 @@ static void
 delay_cb(int fd, int what, void *v)
 {
     struct replay_closure *closure = v;
-    const struct timing_closure *timing = &closure->timing;
+    struct timing_closure *timing = &closure->timing;
     debug_decl(delay_cb, SUDO_DEBUG_UTIL)
 
-    /* Check for window change event and resize as needed. */
-    if (timing->idx == IOFD_TIMING) {
+    switch (timing->event) {
+    case IO_EVENT_WINSIZE:
 	resize_terminal(timing->u.winsize.rows, timing->u.winsize.cols);
-	next_timing_record(closure);
-	debug_return;
+	break;
+    case IO_EVENT_STDIN:
+	if (io_log_files[IOFD_STDIN].enabled)
+	    timing->fd = io_log_files[IOFD_STDIN].fd;
+	break;
+    case IO_EVENT_STDOUT:
+	if (io_log_files[IOFD_STDOUT].enabled)
+	    timing->fd = io_log_files[IOFD_STDOUT].fd;
+	break;
+    case IO_EVENT_STDERR:
+	if (io_log_files[IOFD_STDERR].enabled)
+	    timing->fd = io_log_files[IOFD_STDERR].fd;
+	break;
+    case IO_EVENT_TTYIN:
+	if (io_log_files[IOFD_TTYIN].enabled)
+	    timing->fd = io_log_files[IOFD_TTYIN].fd;
+	break;
+    case IO_EVENT_TTYOUT:
+	if (io_log_files[IOFD_TTYOUT].enabled)
+	    timing->fd = io_log_files[IOFD_TTYOUT].fd;
+	break;
     }
 
-    /* If we are not replaying this stream, just read the next record. */
-    if (timing->idx >= IOFD_MAX || !io_log_files[timing->idx].enabled) {
+    if (timing->fd.v != NULL) {
+	/* If the stream is open, enable the write event. */
+	if (sudo_ev_add(closure->evbase, closure->output_ev, NULL, false) == -1)
+	    sudo_fatal(U_("unable to add event to queue"));
+    } else {
+	/* Not replaying, get the next timing record and continue. */
 	next_timing_record(closure);
-	debug_return;
     }
 
-    /* We are replaying this strean, enable write event. */
-    if (sudo_ev_add(closure->evbase, closure->output_ev, NULL, false) == -1)
-	sudo_fatal(U_("unable to add event to queue"));
-
     debug_return;
 }
 
@@ -953,7 +943,8 @@ signal_cb(int signo, int what, void *v)
 }
 
 static struct replay_closure *
-replay_closure_alloc(struct timeval *max_delay, const char *decimal, bool interactive)
+replay_closure_alloc(struct timespec *max_delay, const char *decimal,
+    bool interactive, bool suspend_wait)
 {
     struct replay_closure *closure;
     debug_decl(replay_closure_alloc, SUDO_DEBUG_UTIL)
@@ -962,6 +953,7 @@ replay_closure_alloc(struct timeval *max_delay, const char *decimal, bool intera
 	debug_return_ptr(NULL);
 
     closure->interactive = interactive;
+    closure->suspend_wait = suspend_wait;
     closure->timing.max_delay = max_delay;
     closure->timing.decimal = decimal;
 
@@ -1034,14 +1026,16 @@ bad:
 }
 
 static int
-replay_session(struct timeval *max_delay, const char *decimal, bool interactive)
+replay_session(struct timespec *max_delay, const char *decimal,
+    bool interactive, bool suspend_wait)
 {
     struct replay_closure *closure;
     int ret = 0;
     debug_decl(replay_session, SUDO_DEBUG_UTIL)
 
     /* Allocate the delay closure and read the first timing record. */
-    closure = replay_closure_alloc(max_delay, decimal, interactive);
+    closure = replay_closure_alloc(max_delay, decimal, interactive,
+	suspend_wait);
     if (read_timing_record(closure) != 0) {
 	ret = 1;
 	goto done;
@@ -1106,7 +1100,7 @@ write_output(int fd, int what, void *v)
     iov[0].iov_len = nbytes;
 
     if (closure->interactive &&
-	(timing->idx == IOFD_STDOUT || timing->idx == IOFD_STDERR)) {
+	(timing->event == IO_EVENT_STDOUT || timing->event == IO_EVENT_STDERR)) {
 	char *nl;
 
 	/*
@@ -1300,7 +1294,7 @@ static bool
 match_expr(struct search_node_list *head, struct log_info *log, bool last_match)
 {
     struct search_node *sn;
-    bool res, matched = last_match;
+    bool res = false, matched = last_match;
     int rc;
     debug_decl(match_expr, SUDO_DEBUG_UTIL)
 
@@ -1316,7 +1310,8 @@ match_expr(struct search_node_list *head, struct log_info *log, bool last_match)
 	    res = strcmp(sn->u.tty, log->tty) == 0;
 	    break;
 	case ST_RUNASGROUP:
-	    res = strcmp(sn->u.runas_group, log->runas_group) == 0;
+	    if (log->runas_group != NULL)
+		res = strcmp(sn->u.runas_group, log->runas_group) == 0;
 	    break;
 	case ST_RUNASUSER:
 	    res = strcmp(sn->u.runas_user, log->runas_user) == 0;
@@ -1351,146 +1346,6 @@ match_expr(struct search_node_list *head, struct log_info *log, bool last_match)
     debug_return_bool(matched);
 }
 
-static struct log_info *
-parse_logfile(char *logfile)
-{
-    FILE *fp;
-    char *buf = NULL, *cp, *ep;
-    const char *errstr;
-    size_t bufsize = 0, cwdsize = 0, cmdsize = 0;
-    struct log_info *li = NULL;
-    debug_decl(parse_logfile, SUDO_DEBUG_UTIL)
-
-    fp = fopen(logfile, "r");
-    if (fp == NULL) {
-	sudo_warn(U_("unable to open %s"), logfile);
-	goto bad;
-    }
-
-    /*
-     * ID file has three lines:
-     *  1) a log info line
-     *  2) cwd
-     *  3) command with args
-     */
-    if ((li = calloc(1, sizeof(*li))) == NULL)
-	sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
-    if (getline(&buf, &bufsize, fp) == -1 ||
-	getline(&li->cwd, &cwdsize, fp) == -1 ||
-	getline(&li->cmd, &cmdsize, fp) == -1) {
-	sudo_warn(U_("%s: invalid log file"), logfile);
-	goto bad;
-    }
-
-    /* Strip the newline from the cwd and command. */
-    li->cwd[strcspn(li->cwd, "\n")] = '\0';
-    li->cmd[strcspn(li->cmd, "\n")] = '\0';
-
-    /*
-     * Crack the log line (rows and cols not present in old versions).
-     *	timestamp:user:runas_user:runas_group:tty:rows:cols
-     * XXX - probably better to use strtok and switch on the state.
-     */
-    buf[strcspn(buf, "\n")] = '\0';
-    cp = buf;
-
-    /* timestamp */
-    if ((ep = strchr(cp, ':')) == NULL) {
-	sudo_warn(U_("%s: time stamp field is missing"), logfile);
-	goto bad;
-    }
-    *ep = '\0';
-    li->tstamp = sizeof(time_t) == 4 ? strtonum(cp, INT_MIN, INT_MAX, &errstr) :
-	strtonum(cp, LLONG_MIN, LLONG_MAX, &errstr);
-    if (errstr != NULL) {
-	sudo_warn(U_("%s: time stamp %s: %s"), logfile, cp, errstr);
-	goto bad;
-    }
-
-    /* user */
-    cp = ep + 1;
-    if ((ep = strchr(cp, ':')) == NULL) {
-	sudo_warn(U_("%s: user field is missing"), logfile);
-	goto bad;
-    }
-    if ((li->user = strndup(cp, (size_t)(ep - cp))) == NULL)
-	sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
-
-    /* runas user */
-    cp = ep + 1;
-    if ((ep = strchr(cp, ':')) == NULL) {
-	sudo_warn(U_("%s: runas user field is missing"), logfile);
-	goto bad;
-    }
-    if ((li->runas_user = strndup(cp, (size_t)(ep - cp))) == NULL)
-	sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
-
-    /* runas group */
-    cp = ep + 1;
-    if ((ep = strchr(cp, ':')) == NULL) {
-	sudo_warn(U_("%s: runas group field is missing"), logfile);
-	goto bad;
-    }
-    if (cp != ep) {
-	if ((li->runas_group = strndup(cp, (size_t)(ep - cp))) == NULL)
-	    sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
-    }
-
-    /* tty, followed by optional rows + columns */
-    cp = ep + 1;
-    if ((ep = strchr(cp, ':')) == NULL) {
-	/* just the tty */
-	if ((li->tty = strdup(cp)) == NULL)
-	    sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
-    } else {
-	/* tty followed by rows + columns */
-	if ((li->tty = strndup(cp, (size_t)(ep - cp))) == NULL)
-	    sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
-	cp = ep + 1;
-	/* need to NULL out separator to use strtonum() */
-	if ((ep = strchr(cp, ':')) != NULL) {
-	    *ep = '\0';
-	}
-	li->rows = strtonum(cp, 1, INT_MAX, &errstr);
-	if (errstr != NULL) {
-	    sudo_debug_printf(SUDO_DEBUG_ERROR|SUDO_DEBUG_LINENO,
-		"%s: tty rows %s: %s", logfile, cp, errstr);
-	}
-	if (ep != NULL) {
-	    cp = ep + 1;
-	    li->cols = strtonum(cp, 1, INT_MAX, &errstr);
-	    if (errstr != NULL) {
-		sudo_debug_printf(SUDO_DEBUG_ERROR|SUDO_DEBUG_LINENO,
-		    "%s: tty cols %s: %s", logfile, cp, errstr);
-	    }
-	}
-    }
-    fclose(fp);
-    free(buf);
-    debug_return_ptr(li);
-
-bad:
-    if (fp != NULL)
-	fclose(fp);
-    free(buf);
-    free_log_info(li);
-    debug_return_ptr(NULL);
-}
-
-static void
-free_log_info(struct log_info *li)
-{
-    if (li != NULL) {
-	free(li->cwd);
-	free(li->user);
-	free(li->runas_user);
-	free(li->runas_group);
-	free(li->tty);
-	free(li->cmd);
-	free(li);
-    }
-}
-
 static int
 list_session(char *logfile, regex_t *re, const char *user, const char *tty)
 {
@@ -1669,7 +1524,7 @@ read_keyboard(int fd, int what, void *v)
 {
     struct replay_closure *closure = v;
     static bool paused = false;
-    struct timeval tv;
+    struct timespec ts;
     ssize_t nread;
     char ch;
     debug_decl(read_keyboard, SUDO_DEBUG_UTIL)
@@ -1698,16 +1553,16 @@ read_keyboard(int fd, int what, void *v)
 	    break;
 	case '<':
 	    speed_factor /= 2;
-            sudo_ev_get_timeleft(closure->delay_ev, &tv);
-            if (sudo_timevalisset(&tv)) {
+            sudo_ev_get_timeleft(closure->delay_ev, &ts);
+            if (sudo_timespecisset(&ts)) {
 		/* Double remaining timeout. */
-		tv.tv_sec *= 2;
-		tv.tv_usec *= 2;
-		if (tv.tv_usec >= 1000000) {
-		    tv.tv_sec++;
-		    tv.tv_usec -= 1000000;
+		ts.tv_sec *= 2;
+		ts.tv_nsec *= 2;
+		if (ts.tv_nsec >= 1000000000) {
+		    ts.tv_sec++;
+		    ts.tv_nsec -= 1000000000;
 		}
-		if (sudo_ev_add(NULL, closure->delay_ev, &tv, false) == -1) {
+		if (sudo_ev_add(NULL, closure->delay_ev, &ts, false) == -1) {
 		    sudo_debug_printf(SUDO_DEBUG_ERROR|SUDO_DEBUG_LINENO,
 			"failed to double remaining delay timeout");
 		}
@@ -1715,14 +1570,14 @@ read_keyboard(int fd, int what, void *v)
 	    break;
 	case '>':
 	    speed_factor *= 2;
-            sudo_ev_get_timeleft(closure->delay_ev, &tv);
-            if (sudo_timevalisset(&tv)) {
+            sudo_ev_get_timeleft(closure->delay_ev, &ts);
+            if (sudo_timespecisset(&ts)) {
 		/* Halve remaining timeout. */
-		if (tv.tv_sec & 1)
-		    tv.tv_usec += 500000;
-		tv.tv_sec /= 2;
-		tv.tv_usec /= 2;
-		if (sudo_ev_add(NULL, closure->delay_ev, &tv, false) == -1) {
+		if (ts.tv_sec & 1)
+		    ts.tv_nsec += 500000000;
+		ts.tv_sec /= 2;
+		ts.tv_nsec /= 2;
+		if (sudo_ev_add(NULL, closure->delay_ev, &ts, false) == -1) {
 		    sudo_debug_printf(SUDO_DEBUG_ERROR|SUDO_DEBUG_LINENO,
 			"failed to halve remaining delay timeout");
 		}
@@ -1743,98 +1598,11 @@ read_keyboard(int fd, int what, void *v)
     debug_return;
 }
 
-/*
- * Parse a timing line, which is formatted as:
- *	index sleep_time num_bytes
- * Where index is IOFD_*, sleep_time is the number of seconds to sleep
- * before writing the data and num_bytes is the number of bytes to output.
- * Returns true on success and false on failure.
- */
-static bool
-parse_timing(const char *buf, double *seconds, struct timing_closure *timing)
-{
-    unsigned long ul;
-    long l;
-    double d, fract = 0;
-    char *cp, *ep;
-    debug_decl(parse_timing, SUDO_DEBUG_UTIL)
-
-    /* Parse index */
-    ul = strtoul(buf, &ep, 10);
-    if (ep == buf || !isspace((unsigned char) *ep))
-	goto bad;
-    if (ul >= IOFD_MAX) {
-	if (ul != 6)
-	    goto bad;
-	/* work around a bug in timing files generated by sudo 1.8.7 */
-	timing_idx_adj = 2;
-    }
-    timing->idx = (int)ul - timing_idx_adj;
-    for (cp = ep + 1; isspace((unsigned char) *cp); cp++)
-	continue;
-
-    /*
-     * Parse number of seconds.  Sudo logs timing data in the C locale
-     * but this may not match the current locale so we cannot use strtod().
-     * Furthermore, sudo < 1.7.4 logged with the user's locale so we need
-     * to be able to parse those logs too.
-     */
-    errno = 0;
-    l = strtol(cp, &ep, 10);
-    if (ep == cp || (*ep != '.' && strncmp(ep, timing->decimal, strlen(timing->decimal)) != 0))
-	goto bad;
-    if (l < 0 || l > INT_MAX || (errno == ERANGE && l == LONG_MAX))
-	goto bad;
-    *seconds = (double)l;
-    cp = ep + (*ep == '.' ? 1 : strlen(timing->decimal));
-    d = 10.0;
-    while (isdigit((unsigned char) *cp)) {
-	fract += (*cp - '0') / d;
-	d *= 10;
-	cp++;
-    }
-    *seconds += fract;
-    while (isspace((unsigned char) *cp))
-	cp++;
-
-    if (timing->idx == IOFD_TIMING) {
-	errno = 0;
-	ul = strtoul(cp, &ep, 10);
-	if (ep == cp || !isspace((unsigned char) *ep))
-	    goto bad;
-	if (ul > INT_MAX || (errno == ERANGE && ul == ULONG_MAX))
-	    goto bad;
-	timing->u.winsize.rows = (int)ul;
-	for (cp = ep + 1; isspace((unsigned char) *cp); cp++)
-	    continue;
-
-	errno = 0;
-	ul = strtoul(cp, &ep, 10);
-	if (ep == cp || *ep != '\0')
-	    goto bad;
-	if (ul > INT_MAX || (errno == ERANGE && ul == ULONG_MAX))
-	    goto bad;
-	timing->u.winsize.cols = (int)ul;
-    } else {
-	errno = 0;
-	ul = strtoul(cp, &ep, 10);
-	if (ep == cp || *ep != '\0')
-	    goto bad;
-	if (ul > SIZE_MAX || (errno == ERANGE && ul == ULONG_MAX))
-	    goto bad;
-	timing->u.nbytes = (size_t)ul;
-    }
-
-    debug_return_bool(true);
-bad:
-    debug_return_bool(false);
-}
-
 static void
 usage(int fatal)
 {
     fprintf(fatal ? stderr : stdout,
-	_("usage: %s [-hnR] [-d dir] [-m num] [-s num] ID\n"),
+	_("usage: %s [-hnRS] [-d dir] [-m num] [-s num] ID\n"),
 	getprogname());
     fprintf(fatal ? stderr : stdout,
 	_("usage: %s [-h] [-d dir] -l [search expression]\n"),
@@ -1854,6 +1622,7 @@ help(void)
 	"  -h, --help           display help message and exit\n"
 	"  -l, --list           list available session IDs, with optional expression\n"
 	"  -m, --max-wait=num   max number of seconds to wait between events\n"
+	"  -S, --suspend-wait   wait while the command was suspended\n"
 	"  -s, --speed=num      speed up or slow down output\n"
 	"  -V, --version        display version information and exit"));
     exit(0);
