@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2009-2017 Todd C. Miller <Todd.Miller@courtesan.com>
+ * Copyright (c) 2009-2017 Todd C. Miller <Todd.Miller@sudo.ws>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -14,6 +14,11 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+/*
+ * This is an open source non-commercial project. Dear PVS-Studio, please check it.
+ * PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
+ */
+
 #include <config.h>
 
 #include <sys/types.h>
@@ -311,6 +316,9 @@ sudo_terminated(struct command_status *cstat)
     for (signo = 0; signo < NSIG; signo++) {
 	if (signal_pending(signo)) {
 	    switch (signo) {
+	    case SIGCHLD:
+		/* Ignore. */
+		break;
 	    case SIGTSTP:
 		/* Suspend below if not terminated. */
 		sigtstp = true;
@@ -377,34 +385,37 @@ sudo_execute(struct command_details *details, struct command_status *cstat)
     }
 
     /*
-     * If we have an I/O plugin or the policy plugin has requested one, we
-     * need to allocate a pty.
+     * Run the command in a new pty if there is an I/O plugin or the policy
+     * has requested a pty.  If /dev/tty is unavailable and no I/O plugin
+     * is configured, this returns false and we run the command without a pty.
      */
     if (!TAILQ_EMPTY(&io_plugins) || ISSET(details->flags, CD_USE_PTY)) {
-	/*
-	 * Run the command in a new pty, wait for it to finish and
-	 * send the plugin the exit status.
-	 */
-	exec_pty(details, cstat);
-    } else if (!ISSET(details->flags, CD_SET_TIMEOUT|CD_SUDOEDIT) &&
+	if (exec_pty(details, cstat))
+	    goto done;
+    }
+
+    /*
+     * If we are not running the command in a pty, we were not invoked
+     * as sudoedit, there is no command timeout and there is no close
+     * function, just exec directly.  Only returns on error.
+     */
+    if (!ISSET(details->flags, CD_SET_TIMEOUT|CD_SUDOEDIT) &&
 	policy_plugin.u.policy->close == NULL) {
-	/*
-	 * If we are not running the command in a pty, we were not invoked
-	 * as sudoedit, there is no command timeout and there is no close
-	 * function, just exec directly.  Only returns on error.
-	 */
 	if (!sudo_terminated(cstat)) {
 	    exec_cmnd(details, -1);
 	    cstat->type = CMD_ERRNO;
 	    cstat->val = errno;
 	}
-    } else {
-	/*
-	 * No pty but we need to wait for the command to finish to
-	 * send the plugin the exit status.
-	 */
-	exec_nopty(details, cstat);
+	goto done;
     }
+
+    /*
+     * Run the command in the existing tty (if any) and wait for it to finish.
+     */
+    exec_nopty(details, cstat);
+
+done:
+    /* The caller will run any plugin close functions. */
     debug_return_int(cstat->type == CMD_ERRNO ? -1 : 0);
 }
 
